# -*- coding: utf-8 -*-
import re

__author__ = 'luckydonald'


REPO_NAME_REGEX = re.compile("([a-zA-Z0-9_.-]+)")

AT_USERNAME_REGEX = re.compile(  # https://regex101.com/r/eJ6cA4/
    "(?:\s|^|\.|,|:|\?|!|¡|¿)@(?!-)(?P<user>(?:[a-zA-Z0-9]|-(?!-))+[a-zA-Z0-9])(?![^\x00-\x7F])(?:$|[^a-zA-Z0-9\-_])"
)

FILE_URL_REGEX = re.compile(  # https://regex101.com/r/qK6uN9
    "\\b(?P<url>https?://github.com/(?P<user>[a-zA-Z0-9_-]+)/(?P<repo>[a-zA-Z0-9_.-]+)/?"
    "(?P<path>(?:(?P<kind>blob|tree)/(?P<branch>[a-zA-Z0-9_%-]+))/?(?P<file>[^\s#]*)/?)?)"
    "(?:#(?P<hash>[-;/?:@&=+$,_.!~*'()%0-9a-zA-Z]*))?(?:\\s|$)", flags=re.IGNORECASE
)
"""
Matching groups:
  - url: the complete url
      - user: git user or organisation
      - repo: the repository
      - path: When existent, this is not the project page (root of master)
          - kind: blob or tree
          - branch: the name of the branch (kind=tree), or the commit hash (kind=blob)
        - file: the rest of the filepath (from root of that branch, can be empty)
      - hash: Can be non-existent or empty. Everything behind the #.
"""






# -*- coding: utf-8 -*-
import logging as _logging
import sys

__author__ = 'luckydonald'
__all__ = ["logging", "ColoredFormatter"]


class ColoredFormatter(_logging.Formatter):
    class Color(object):
        """
         utility to return ansi colored text.
         just to store the colors next to the function.
        """
        # Color codes: http://misc.flogisoft.com/bash/tip_colors_and_formatting

        colors = {
            'default': 39,
            'black': 30,
            'red': 31,
            'green': 32,
            'yellow': 33,
            'blue': 34,
            'magenta': 35,
            'cyan': 36,
            'white': 37,
            'grey': 90,
            'bgred': 41,
            'bggrey': 100
        }
        mapping = {
            'INFO': 'default',
            'WARNING': 'yellow',
            'ERROR': 'red',
            'CRITICAL': 'magenta',
            'DEBUG': 'grey',
            'SUCCESS': 'green'
        }

        color_prefix = '\033['

        def prepare_color(self, color_number):
            return ('%s%dm') % (self.color_prefix, color_number)

        # end def

        def colored(self, record):
            """
            adsasd
            """

            color = self.mapping.get(record.levelname, 'default')
            clr = self.colors[color]

            filepart = record.treadName + ": " if hasattr(record,
                                                          "treadName") and record.treadName != "MainTread" else ""
            filepart += record.name if record.name else ""
            filepart += "." + record.funcName + ": " if record.funcName != "<module>" else ": "
            filepart = filepart.join(
                [self.prepare_color(94), self.prepare_color(39)])  # Light blue ,  Default foreground color
            formatter = dict(
                all_off=self.prepare_color(0),  # Reset all attributes
                color_on=self.prepare_color(clr),  # Color as given/from lookup
                color_off=self.prepare_color(39),  # Default foreground color
                inverse_on=self.prepare_color(7),  # Reverse (invert the foreground and background colors)
                inverse_off=self.prepare_color(27),  # Reset reverse
                background_off=self.prepare_color(49),  # Default background color)
            )
            lines = []
            level = "{:8}".format(record.levelname)
            level_filler = "{:{}}".format("", len(level))
            lines_ = record.message.splitlines()
            first_line = True if len(lines_) > 1 else  None
            for line in lines_:
                if first_line is None:  # single line
                    lines.append(
                        "{color_on}{inverse_on}{level}{inverse_off}{color_off} {filepart}{color_on}{message}{color_off}{background_off}{all_off}".format(
                            filepart=filepart, level=level, message=line, **formatter))
                    break
                elif first_line:  # first line
                    lines.append("{color_on}{inverse_on}{level}{inverse_off}{color_off} {filepart}{all_off}".format(
                        filepart=filepart, level=level, message=line, **formatter))
                lines.append(
                    "{color_on}{inverse_on}{level_filler}{inverse_off}{color_off} {color_on}{message}{color_off}{background_off}{all_off}".format(
                        level_filler=level_filler, message=line, **formatter))
                first_line = False
            # end for
            return "\n".join(lines)
            # end def

    colored = Color().colored

    # firstpart = _logging.Formatter("%(levelname)s: %(threadName)s %(name)s.%(funcName)s: %(message)s")
    def format(self, record):
        super(ColoredFormatter, self).format(record)
        # if record.threadName == "MainThread":
        # 	pass
        # part1 = self.firstpart.format(record)
        s = self._fmt % record.__dict__  # py3: s = self.formatMessage(record)
        if record.exc_text:
            if s[-1:] != "\n":
                s += "\n"
            try:
                s = s + record.exc_text
            except UnicodeError:  # PYTHON 2, LOL!
                # Sometimes filenames have non-ASCII chars, which can lead
                # to errors when s is Unicode and record.exc_text is str
                # See issue 8924.
                # We also use replace for when there are multiple
                # encodings, e.g. UTF-8 for the filesystem and latin-1
                # for a script. See issue 13232.
                s = s + record.exc_text.decode(sys.getfilesystemencoding(), 'replace')
        if hasattr(record, "stack_info") and record.stack_info:  # py2 doesn't have .stack_info
            if s[-1:] != "\n":
                s += "\n"
            s = s + record.stack_info  # py3: self.formatStack()
        record.message = s
        return self.colored(record)


# noinspection PyProtectedMember,PyProtectedMember
class _LoggingWrapper(object):
    SUCCESS = 25  # between WARNING and INFO

    def __init__(self):
        _logging.addLevelName(self.SUCCESS, 'SUCCESS')

    def getLoglevelInt(self, level_string):
        """
        You provide a String, and get a level int
        :param level_string: The level.
        :type  level_string: str
        :return: level
        :rtype : int
        :raises KeyError: if the level does not exists.
        """
        return {
            "NOTSET": _logging.NOTSET,
            "DEBUG": _logging.DEBUG,
            "INFO": _logging.INFO,
            "SUCCESS": self.SUCCESS,
            "WARNING": _logging.WARNING,
            "WARN": _logging.WARN,  # = WARNING
            "ERROR": _logging.ERROR,
            "FATAL": _logging.FATAL,  # = CRITICAL
            "CRITICAL": _logging.CRITICAL,
        }[level_string]

    def __call__(self, logger_name):
        """
        alias to logger.getLogger(logger_name)
        :param logger_name:
        :return: self.getLogger(logger_name)
        """
        return self.getLogger(logger_name)

    def add_colored_handler(self, logger_name=None, stream=None, level=None):
        """
        Register a logger handler to colorfull print the messages.

        If stream is specified, the instance will use it for logging output; otherwise, sys.stdout will be used.

        :param logger_name: the name of the logger you want to register the printing to.
        Probably you should use __name__ , to use your package's logger, "root" will force all loggers to output. DO NOT use this in librarys.

        :param stream: An output stream. Default: sys.stdout

        :return: None
        """
        logger = self.getLogger(logger_name)  # __name__
        if stream is None:
            import sys
            stream = sys.stdout
        handler = _logging.StreamHandler(stream=stream)
        formatter = ColoredFormatter()
        handler.setFormatter(formatter)
        logger.addHandler(handler)
        if level:
            logger.setLevel(level)
        return logger

    def test_logger_levels(self):
        logger = self.getLogger(__name__)
        logger_level = logger.getEffectiveLevel()
        logger.setLevel(logging.DEBUG)
        logger.debug('level debug')
        logger.info('level info')
        logger.success('level success')
        logger.warning('level warning')
        logger.error('level error')
        logger.critical('level critical')
        logger.setLevel(logger_level)

    def getLogger(self, name=None):
        """
        Adds the .success() function to the logger, else it is same as logger.getLogger()
        :param logger: a logging.getLogger() logger.
        :return:
        """
        logger = _logging.getLogger(name)
        logger.SUCCESS = self.SUCCESS
        setattr(logger, "success", lambda message, *args: logger._log(self.SUCCESS, message, args))
        return logger

    if sys.version < "3":
        def success(self, msg, *args, **kwargs):
            """
            Log 'msg % args' with severity 'SUCCESS'.

            To pass exception information, use the keyword argument exc_info with
            a true value.

            logger.debug("Houston, we landed in the %s", "moon", exc_info=False)
            """
            self._success(msg, *args, **kwargs)
    else:
        from .py3 import success

    def _success(self, msg, *args, **kwargs):
        if len(self.root.handlers) == 0:
            self.basicConfig()
        self.root._log(self.SUCCESS, msg, args, **kwargs)

    def __getattr__(self, item):
        if item != "__getattr__":
            if item in self.__dict__:
                return self.__dict__[item]
        if item == "getLogger":
            return self.getLogger
        elif item == "success":
            return self.success
        elif item == "SUCCESS":
            return self.SUCCESS
            # end if
            pass
        else:
            return getattr(_logging, item)


logging = _LoggingWrapper()



"""
    Only one Instance of the class.
    Like in Java .getInstance()

    Python 3:
    apply with keyword argument in the base-class list:
    >>> class Example(object, metaclass = Singleton):
    >>>    pass

    Python 2.7:
    apply like
    >>> class Exampe(object):
    >>>    __metaclass__ = Singleton
    >>>    pass
"""


class Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]



# -*- coding: utf-8 -*-
from .dependencies import import_or_install

import tempfile
import mimetypes  # get mime types/suffix for DL.

import os
import errno  # exist_ok workaround

# open_folder()
import subprocess
import sys
import logging
magic = import_or_install("magic", "python-magic")  # pip install python-magic

__author__ = 'luckydonald'

logger = logging.getLogger(__name__)


def gettempdir(temp_folder_name="luckydonald-utils"):
    temp_dir = tempfile.gettempdir()
    temp_dir = os.path.join(temp_dir, temp_folder_name)
    # py3
    # os.makedirs(temp_dir, exist_ok=True) #don't raise errors if existent.
    # py2/3 exist_ok workaround
    try:
        os.makedirs(temp_dir)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise
    # end exist_ok workaround
    return temp_dir


# end def


def get_file_suffix(file_path=None, file_url=None):
    """
    This calls `get_file_mime()` to get the mime, and then calls `guess_extension()`.
    """
    mime = get_file_mime(file_path=file_path, file_url=file_url)
    return guess_extension(mime)


def guess_extension(mime):
    return mimetypes.guess_extension(type=mime or "")


def get_file_mime(file_path=None, file_url=None):
    if file_url:
        url = file_url
    elif file_path:
        from urllib.request import pathname2url
        url = pathname2url(file_path)
    else:
        raise AttributeError("Neither URL (file_url) nor local path (file_path) given.")
    mime = magic.from_file(file_path).decode("utf-8")
    return mime


def get_byte_mime(bytes):
    magic.from_buffer(bytes, mime=True).decode("utf-8")


def do_a_filename(input_file_name):
    output_file_name = input_file_name
    for replacer in [(":", ""), ("?", ""), ("*", ""), ("”", "\""), (">", ""), ("<", ""), ("|", "-"), ("\\", " "),
                     ("/", " ")]:
        output_file_name = output_file_name.replace(replacer[0], replacer[1])
    logger.debug("Filename '{old_filename}' is now '{new_filename}'.".format(old_filename=input_file_name,
                                                                             new_filename=output_file_name))
    return output_file_name


"""
def open_folder(folder_path)
"""  # for different Platforms
if sys.platform == 'darwin':  # Mac OS
    def open_folder(folder_path):
        subprocess.check_call(['open', '--', folder_path])


    def open_file_folder(file_path):
        subprocess.check_call(['open', '-R', '--', file_path])
elif sys.platform == 'linux2':  # linux, hopefully has gnome?
    def open_folder(folder_path):
        subprocess.check_call(['gnome-open', '--', folder_path])  # untested.


    def open_file_folder(file_path):
        open_folder(os.path.dirname(file_path))
elif sys.platform == 'win32':  # windows
    def open_folder(folder_path):
        subprocess.check_call(['explorer', folder_path])


    def open_file_folder(file_path):
        try:
            subprocess.check_call(['explorer', '/select,"%s"' % file_path])
        except:
            open_folder(os.path.dirname(file_path))



import functools
import inspect
from DictObject import DictObject
import logging
logger = logging.getLogger(__name__)

__author__ = 'luckydonald'


def caller(func):
    """
    functions decorated with this will be called with an `call` kwarg, containing information about the function itself, and the caller.
    If the caller could not be fetched correctly, the `caller`s attributes all will be `None`.
    """

    @functools.wraps(func)
    def new_func(*args, **kwargs):
        try:
            stack = inspect.stack()[1]
            call = {"self": {"name": func.__name__, "file": func.__code__.co_filename,
                             "line": func.__code__.co_firstlineno + 1},
                    "caller": {"name": stack[3], "file": stack[1], "line": stack[2], "code": stack[4][0]}
                    }
        except TypeError:
            call = {"self": {"name": func.__name__, "file": func.__code__.co_filename,
                             "line": func.__code__.co_firstlineno + 1},
                    "caller": {"name": None, "file": None, "line": None, "code": None}
                    }
        kwargs["call"] = call
        return func(*args, **kwargs)

    return new_func


def deprecated(func):
    """This is a decorator which can be used to mark functions
    as deprecated. It will result in a warning being logged
    when the function is used."""

    @functools.wraps(func)
    def new_func(*args, **kwargs):
        try:
            logger.warn(
                "Call to deprecated function {} at file {}:{}. Called from function {} at file {}:{}\nLine: {}".format(
                    func.__name__, func.__code__.co_filename,  # python2: func.func_code.co_filename
                    func.__code__.co_firstlineno + 1,  # python2: func.func_code.co_firstlineno
                    inspect.stack()[1][3], inspect.stack()[1][1], inspect.stack()[1][2], inspect.stack()[1][4][0],
                ))
        except TypeError:
            logger.warn(
                "Call to deprecated function {} at file {}:{}.".format(func.__name__, func.__code__.co_filename,
                                                                       func.__code__.co_firstlineno + 1))
        return func(*args, **kwargs)

    return new_func


def gone(func):
    """This is a decorator which can be used to mark functions
    as deprecated, and gone. It will raise a NotImplementedError
    when the function is used."""
    @functools.wraps(func)
    def new_func(*args, **kwargs):
        raise NotImplementedError(
            "Call to deprecated function {} at file {}:{}. Called from function {} at file {}:{}\nLine: {}".format(
                func.__name__, func.__code__.co_filename,  # python2: func.func_code.co_filename
                func.__code__.co_firstlineno + 1,  # python2: func.func_code.co_firstlineno
                inspect.stack()[1][3], inspect.stack()[1][1], inspect.stack()[1][2], inspect.stack()[1][4][0]
            ))

    return new_func



# -*- coding: utf-8 -*-
__author__ = 'luckydonald'

CHARS_UNESCAPED = ["\\", "\n", "\r", "\t", "\b", "\a", "'"]
CHARS_ESCAPED = ["\\\\", "\\n", "\\r", "\\t", "\\b", "\\a", "\\'"]


def text_split(text, limit, max_parts=None):
    """
    Split a text usefull.
    Prefer splitting at linebreaks, sentences, words.

    :type text: str
    :type limit: int
    :param max_parts: If not None: In how many parts it should be splitted. The last element might be longer as the given limit.
    :type  max_parts: None | int
    :return:
    """
    if not isinstance(text, str):
        raise TypeError("text is not str but {}".format(type(text)))
    assert isinstance(limit, int)
    assert max_parts is None or isinstance(max_parts, int)
    text = text.strip()
    if len(text) == 0 or limit == 0:
        return []
    if max_parts == 1:
        return [text]
    assert max_parts is None or max_parts > 1
    # LinkRanges lnkRanges = textParseLinks(text, TextParseLinks | TextParseMentions | TextParseHashtags);
    # currentLink = 0
    # lnkCount = lnkRanges.size()
    half = limit // 2
    good_level = 0
    good = 0
    end = len(text)
    unicode_i = 0
    for i in range(0, end):
        char = text[i]
        # while (currentLink < lnkCount && ch >= lnkRanges[currentLink].from + lnkRanges[currentLink].len) {
        # 	++currentLink;
        #  }
        # bool inLink = (currentLink < lnkCount) && (ch > lnkRanges[currentLink].from) &&
        #   (ch < lnkRanges[currentLink].from + lnkRanges[currentLink].len);
        if unicode_i > half:
            if char == "\n":
                if i + 1 < end and text[i + 1] == "\n" and good_level <= 7:
                    good_level = 7
                    good = i
                elif good_level <= 6:
                    good_level = 6
                    good = i
            elif char.isspace():
                if is_sentence_end(char) and good_level <= 5:
                    good_level = 5
                    good = i
                elif is_sentence_part_end(text[i - 1]) and good_level <= 4:
                    good_level = 4
                    good = i
                elif good_level <= 3:
                    good_level = 3
                    good = i
            elif is_word_separator(text[i - 1]) and good_level <= 2:
                good_level = 2
                good = i
            elif good_level <= 1:
                good_level = 1
                good = i
                # end if
        # end if
        if unicode_i >= limit:
            if max_parts == 1:
                return [text.strip()]
            first_text = text[:good]
            text = text[good:]
            parts = [first_text.strip(), ]
            if max_parts is None:
                parts.extend(text_split(text, limit))
            elif max_parts > 1:
                parts.extend(text_split(text, limit, max_parts=max_parts - 1))
            else:
                raise ArithmeticError("max_split reached illegal state: {}".format(max_parts))
            return parts
        # end if
        if char in CHARS_UNESCAPED:
            unicode_i += 2
        else:
            unicode_i += len(char.encode('utf-8'))
    # end for
    return [text]


# end def
"""
        if len(text)> MAX_TEXT_LENGTH:
            last_space = text.rfind(" ", MAX_TEXT_LENGTH-40, MAX_TEXT_LENGTH)
            last_newline = text.rfind("\n", MAX_TEXT_LENGTH-40, MAX_TEXT_LENGTH)
            if abs(last_newline-last_space) < 20:
                split_at = last_newline
            else:
                split_at = last_space
            assert split_at <= MAX_TEXT_LENGTH
            self.text = text[:split_at]
            next_text = text[split_at+1:] # omit the space or newline.
            logger.debug("Splitted {num} character long message text to {len_a} and {len_b}.".format(
                num=len(text), len_a=len(self.text), len_b=len(next_text)))
            self._next_msg = TextMessage(next_text, receiver)
"""


def is_sentence_end(char):
    return char in ".?!"


def is_sentence_part_end(char):
    return char in ",:;"


def is_word_separator(char):
    return char in [" ", "\n", ".", ",", "?", "!", "@", "#", "$", ":", ";", "-", "<", ">", "[", "]", "(", ")", "{", "}",
                    "=", "/", "+", "%", "&", "^", "*", "'", "\"", "`", "~", "|"]


def escape(string):
    for i in range(0, 7):
        string = string.replace(CHARS_UNESCAPED[i], CHARS_ESCAPED[i])
    return string
